package Bitmap;

import java.awt.*;
import java.util.*;
import LinearAlgebra.*;
import Paintable.*;

public class Template extends BitmapClass {
    private final int LABEL_FINAL = 1024;
    private final int LABEL_CHECKED = 1023;
    private final int LABEL_MARKED = 1000;
    private final int [][] neighbors = {{-1,-1},{-1,0},{-1,1},{0,1},{1,1},{1,0},{1,-1},{0,-1}};
    private Graph initializerGraph;

    public Template() {
	super();
    }

    public Template(BitmapClass bitmap) {
	super(bitmap);
    }

    public Template(int in_xSize,int in_ySize) {
	super(in_xSize,in_ySize);
    }

    public BitmapClass Clone() {
	BitmapClass copy = super.Clone();
	return new Template(copy);
    }

    // Starts at the left neighbor, goes clockwise
    // \cite{ArSa85}
    final private int neighborLabel(int i,int j,int k) {
	return GetPixelExtended(i + neighbors[k%8][0],j + neighbors[k%8][1]);
    }

    final private boolean neighbor(int i,int j,int k) {
	return GetPixelExtended(i + neighbors[k%8][0],j + neighbors[k%8][1],0);
    }

    // min(n5+1,n6+1,n7+1,n8+1)
    // \cite{ArSa85}    
    final private int f1EightDistance(int i,int j) {
	int n5 = neighborLabel(i,j,5),n6 = neighborLabel(i,j,6),n7 = neighborLabel(i,j,7),n8 = neighborLabel(i,j,8);
	if (n7 <= n8)
	    if (n6 <= n7)
		if (n5 <= n6)
		    return n5 + 1;
		else
		    return n6 + 1;
	    else
		if (n5 <= n7)
		    return n5 + 1;
		else
		    return n7 + 1;
	else
	    if (n6 <= n8)
		if (n5 <= n6)
		    return n5 + 1;
		else
		    return n6 + 1;
	    else
		if (n5 <= n8)
		    return n5 + 1;
		else
		    return n8 + 1;
    }

    // min(n1+1,n2+1,n3+1,n4+1)
    // \cite{ArSa85}
    final private int f2EightDistance(int i,int j) {
	int n1 = neighborLabel(i,j,1),n2 = neighborLabel(i,j,2),n3 = neighborLabel(i,j,3),n4 = neighborLabel(i,j,4),n;
	if (n3 <= n4)
	    if (n2 <= n3)
		if (n1 <= n2)
		    n = n1 + 1;
		else
		    n = n2 + 1;
	    else
		if (n1 <= n3)
		    n = n1 + 1;
		else
		    n = n3 + 1;
	else
	    if (n2 <= n4)
		if (n1 <= n2)
		    n = n1 + 1;
		else
		    n = n2 + 1;
	    else
		if (n1 <= n4)
		    n = n1 + 1;
		else
		    n = n4 + 1;
	int p = GetPixel(i,j);
	if (n > p)
	    return p;
	else
	    return n;
    }

    // min(n5+1,n7+1)
    final private int f1FourDistance(int i,int j) {
	int n5 = neighborLabel(i,j,5),n7 = neighborLabel(i,j,7);
	if (n5 < n7)
	    return n5 + 1;
	else
	    return n7 + 1;
    }

    // min(n1+1,n2+1,n3+1,n4+1)
    final private int f2FourDistance(int i,int j) {
	int n1 = neighborLabel(i,j,1),n3 = neighborLabel(i,j,3),n;
	if (n1 < n3)
	    n = n1 + 1;
	else
	    n = n3 + 1;
	int p = GetPixel(i,j);
	if (n > p)
	    return p;
	else
	    return n;
    }

    // \cite{ArSa85}
    final private int eightConnectivity(int i,int j) {
	boolean n1 = neighbor(i,j,1),n2 = neighbor(i,j,2),n3 = neighbor(i,j,3),n4 = neighbor(i,j,4);
	boolean n5 = neighbor(i,j,5),n6 = neighbor(i,j,6),n7 = neighbor(i,j,7),n8 = neighbor(i,j,8);
	int c8 = 0;
	if (!n1 && (n2 || n3))
	    c8++;
	if (!n3 && (n4 || n5))
	    c8++;
	if (!n5 && (n6 || n7))
	    c8++;
	if (!n7 && (n8 || n1))
	    c8++;
	return c8;
    }
    
    // \cite{SuAb86}
    final private boolean a(int i,int j) {
	return eightConnectivity(i,j) == 1;
    }
    
    // \cite{SuAb86}
    final private int NumOfNeighbors(int i,int j) {
	int n = 0;
	for(int k=1;k<=8;k++)
	    if (neighbor(i,j,k))
		n++;
	return n;
    }
    final private boolean b(int i,int j) {
	return NumOfNeighbors(i,j) > 1;
    }
    
    final private boolean c(int i,int j) {
	int n1 = neighborLabel(i,j,1),n3 = neighborLabel(i,j,3),
	    n5 = neighborLabel(i,j,5),n7 = neighborLabel(i,j,7),p = GetPixel(i,j);
	if (n1 > 0 && p - n1 == 1)
	    return false;
	if (n3 > 0 && p - n3 == 1)
	    return false;
	if (n5 > 0 && p - n5 == 1)
	    return false;
	if (n7 > 0 && p - n7 == 1)
	    return false;
	return true;
    }
    
    final private boolean finalPixel(int i,int j) {
	return GetPixel(i,j) == LABEL_FINAL;
    }

    final private void setFinalPixel(int i,int j) {
	SetPixel(i,j,LABEL_FINAL);
    }

    final private void setCheckedPixel(int i,int j) {
	SetPixel(i,j,LABEL_CHECKED);
    }

    // \cite{SuAb86}
    final private boolean d(int i,int j) {
	return !finalPixel(i,j);
    }

    final private void deletePixel(int i,int j) {
	SetPixel(i,j,0);
    }

    final private boolean borderPixel(int i,int j) {
	if (GetPixel(i,j,0) && (!neighbor(i,j,1) || !neighbor(i,j,3) || !neighbor(i,j,5) || !neighbor(i,j,7)))
	    return true;
	else
	    return false;
    }

    int labelToTrace;

    // \cite{SuAb86}
    final public boolean DeletePixels() {
	int [][] borderPoints = new int[GetXSize()*GetYSize()][2];
	int numOfborderPoints = 0;
	labelToTrace++;
//	System.out.println("labelToTrace = " + labelToTrace);
	int pj=GetYSize() - 1;
	int i=0,j=0,p;
	boolean moreToDelete = true;
	boolean changed = false;
	while(moreToDelete) {
	    moreToDelete = false;
	    for(j=pj;j>=0 && !moreToDelete;j--)
		for(i=0;i<GetXSize() && !moreToDelete;i++) {
		    p = GetPixel(i,j);
		    if (p == labelToTrace)
			moreToDelete = true;
		}
	    pj = ++j;
	    --i;
	    borderPoints[numOfborderPoints][0] = i;
	    borderPoints[numOfborderPoints][1] = j;
	    numOfborderPoints++;
	    if (moreToDelete) {
//		int labelToTrace = GetPixel(i,j) - 1;
		boolean moreToPeel = true;
		boolean justFinished = false;
		while(moreToPeel) {
		    moreToPeel = false;
//		    labelToTrace++;
		    while(numOfborderPoints > 0) {
			numOfborderPoints--;
			i = borderPoints[numOfborderPoints][0];
			j = borderPoints[numOfborderPoints][1];
 			for(int k=1;k<=8;k++) {
			    if (neighborLabel(i,j,k) == labelToTrace && 
				borderPixel(i + neighbors[k%8][0],j + neighbors[k%8][1])) {
				borderPoints[numOfborderPoints][0] = i + neighbors[k%8][0];
				borderPoints[numOfborderPoints][1] = j + neighbors[k%8][1];
				numOfborderPoints++;
			    }
			}    
			if (!justFinished) {
//			    System.out.println("pixel = (" + i + "," + j + "),label = " + GetPixel(i,j));
			    if (!a(i,j) || !b(i,j))
				setFinalPixel(i,j);
			    else if (c(i,j) && d(i,j)) {
				deletePixel(i,j);
				changed = true;
			    }
			}
		    }
		}
	    }
	}
	return changed;
    }
    
    // \cite{ArSa85}
    final public void EightDistanceTransformation() {
	for(int j=0;j<GetYSize();j++)
	    for(int i=GetXSize() - 1;i>=0;i--)
		if (GetPixel(i,j,0))
		    SetPixel(i,j,f1EightDistance(i,j));		
	for(int j=GetYSize() - 1;j>=0;j--)
	    for(int i=0;i<GetXSize();i++)
		if (GetPixel(i,j,0))
		    SetPixel(i,j,f2EightDistance(i,j));		
    }
	    	    
    final public void FourDistanceTransformation() {
	for(int j=0;j<GetYSize();j++)
	    for(int i=GetXSize() - 1;i>=0;i--)
		if (GetPixel(i,j,0))
		    SetPixel(i,j,f1FourDistance(i,j));		
	for(int j=GetYSize() - 1;j>=0;j--)
	    for(int i=0;i<GetXSize();i++)
		if (GetPixel(i,j,0))
		    SetPixel(i,j,f2FourDistance(i,j));		
    }
	    	    
    final public void Skeletonize() {
	FourDistanceTransformation();
	labelToTrace = 0;
	while(DeletePixels());
    }

    // The label of each vertex is its degree
    final public void SetInitializerGraph() {
	Skeletonize();
	initializerGraph = new Graph();
	Stack vertexStack = new Stack();
	Stack joinVertexStack = new Stack();
	Stack lineVertexStack = new Stack();
	int i = 0,j = 0;
	VektorDDLabeled vertex,newVertex;
	// Each label = number of neighbors + LABEL_FINAL
	for(j=0;j<GetYSize();j++)
	    for(i=GetXSize() - 1;i>=0;i--)
		if (GetPixel(i,j) > 0)
		    SetPixel(i,j,NumOfNeighbors(i,j) + LABEL_FINAL);
	boolean foundVertex = true;
	// for all connected components
	int iNew,jNew,n,nLine,nLabel,joinVertexIndex;
	boolean found;
	while(foundVertex) {
	    // Find an unprocessed initial line vertex with two line (or end) vertex neighbors or end vertex
	    // We assume that Salt & Pepper noise has been removed, so there are no black 
	    // vertices with zero black neighbors
	    boolean foundLineVertex = false;
	    boolean foundEndVertex = false;
	    int iEndVertex = 0,jEndVertex = 0;
	    for(j=0;j<GetYSize() && !foundLineVertex;j++) {
		for(i=GetXSize() - 1;i>=0 && !foundLineVertex;i--) {
		    if (GetPixel(i,j) == 2 + LABEL_FINAL) {
			foundLineVertex = true;
			for(int k=1;k<=8 && foundLineVertex;k++) {
			    nLabel = neighborLabel(i,j,k);
			    foundLineVertex = foundLineVertex && neighborLabel(i,j,k) <= 2 + LABEL_FINAL;
			}
		    }
		    else if (!foundEndVertex && GetPixel(i,j) == 1 + LABEL_FINAL) {
			iEndVertex = i;
			jEndVertex = j;
			foundEndVertex = true;
		    }
		}
	    }	    
	    foundVertex = foundLineVertex || foundEndVertex;
	    boolean first = true;
	    if (foundLineVertex) {
		// Depth first walk through the skeleton. 
		i++;j--;
		vertexStack.push(new Pixel(i,j));
		vertex = ConvertPixelToVektorDDLabeled(i,j);
		vertex.SetLabel(2);
		initializerGraph.AddPoint(vertex);
		Pixel pixel;
		while(!vertexStack.empty()) {
		    pixel = (Pixel)vertexStack.pop();
		    i = pixel.i;
		    j = pixel.j;
		    vertex = ConvertPixelToVektorDDLabeled(i,j);
		    // if end vertex, just label it LABEL_MARKED
		    if (GetPixel(i,j) == 1 + LABEL_FINAL) {
			SetPixel(i,j,LABEL_MARKED);
		    }
		    // if line vertex
 		    if (GetPixel(i,j) == 2 + LABEL_FINAL) {
			int label = LABEL_MARKED;
			found = false;
			// search for an unmarked neighbor, push it to the stack, add the new vertex and edge
			for(int k=1;k<=8;k++) {
			    nLabel = neighborLabel(i,j,k);
			    if (nLabel > 0 && nLabel != LABEL_MARKED) {
				found = true;
				iNew = i + neighbors[k%8][0];
				jNew = j + neighbors[k%8][1];		    
				newVertex = ConvertPixelToVektorDDLabeled(iNew,jNew);
				newVertex.SetLabel(nLabel - LABEL_FINAL);
				if (initializerGraph.FindPoint(newVertex) == -1) {
				    vertexStack.push(new Pixel(iNew,jNew));
				    initializerGraph.AddPoint(newVertex);
				}
				initializerGraph.AddEdge(vertex,newVertex);
			    }
			}
			SetPixel(i,j,label);				
		    }
		    // if join vertex
		    else if (GetPixel(i,j) > 2 + LABEL_FINAL) {
			// push it to the stack of join vertices, initialize vertex to 0
			pixel = new Pixel(i,j);			
			joinVertexStack.push(pixel);
			joinVertexIndex = initializerGraph.FindPoint(vertex);
			double [] coords = {0,0};
			vertex = new VektorDDLabeled(coords,0);
			// push all join vertices in the cluster on the stack of join vertices, and all neighbor
			// line vertices on the stack of line vertices, update vertex to the center of mass
			// of all joinvertices
			n = nLine = 0;
			while(!joinVertexStack.empty()) {
			    pixel = (Pixel)joinVertexStack.pop();
			    i = pixel.i;
			    j = pixel.j;
			    vertex.AddEqual(ConvertPixelToVektorDDLabeled(i,j));
			    n++;
			    for(int k=1;k<=8;k++) {
				nLabel = neighborLabel(i,j,k);
				// end vertex or line vertex
				pixel = new Pixel(i + neighbors[k%8][0],j + neighbors[k%8][1]);
				if (nLabel == 1  + LABEL_FINAL || nLabel == 2 + LABEL_FINAL) {
				    if (lineVertexStack.indexOf(pixel) == -1) {
					nLine++;
					lineVertexStack.push(pixel);
				    }
				    else // the LineVertex is connected twice to the same join
					SetPixel(pixel.i,pixel.j,LABEL_FINAL + 1);
				}
				// another join vertex
				else if (nLabel > 2  + LABEL_FINAL && nLabel <= 8 + LABEL_FINAL && 
					 joinVertexStack.indexOf(pixel) == -1) {
				    joinVertexStack.push(pixel);
				}
			    }
			    SetPixel(i,j,LABEL_MARKED);
			}
			vertex.DivEqual(n);
			vertex.SetLabel(nLine + 1);
			initializerGraph.SetPointAt(vertex,joinVertexIndex);
			// add saved line vertices, and edges, push line vertices to the stack
			while(!lineVertexStack.empty()) {
			    pixel = (Pixel)lineVertexStack.pop();
			    i = pixel.i;
			    j = pixel.j;
			    newVertex = ConvertPixelToVektorDDLabeled(i,j);
			    newVertex.SetLabel(GetPixel(i,j) - LABEL_FINAL);
			    if (initializerGraph.FindPoint(newVertex) == -1) {
				vertexStack.push(new Pixel(i,j));
				initializerGraph.AddPoint(newVertex);
			    }
			    initializerGraph.AddEdge(vertex,newVertex);
			}
		    }
		}
	    }
	    else if (foundEndVertex) {
		// We found an unprocessed end vertex and no unprocessed line vertex means that 
		// this endvertex has another end vertex or join vertex neighbor. In either case,
		// we create one edge with the two vertices as endpoints.
		i = iEndVertex;
		j = jEndVertex;
		SetPixel(i,j,LABEL_MARKED);
		vertex = ConvertPixelToVektorDDLabeled(i,j);
		vertex.SetLabel(1);
		initializerGraph.AddPoint(vertex);
		for(int k=1;k<=8;k++) {
		    if (neighborLabel(i,j,k) >= 1 + LABEL_FINAL) {
			iNew = i + neighbors[k%8][0];
			jNew = j + neighbors[k%8][1];		    
			newVertex = ConvertPixelToVektorDDLabeled(iNew,jNew);
			SetPixel(iNew,jNew,LABEL_MARKED);
			newVertex.SetLabel(1);
			if (initializerGraph.FindPoint(newVertex) == -1)
			    initializerGraph.AddPoint(newVertex);
			initializerGraph.AddEdge(vertex,newVertex);
			k = 9;
		    }
		}
	    }
	}
	for(j=0;j<GetYSize();j++) {
	    for(i=GetXSize() - 1;i>=0;i--) {
		if (GetPixel(i,j) > 0 && GetPixel(i,j) != LABEL_MARKED) {
		    System.out.println(i + " " + j);
		    throw new RuntimeException("CAN'T INITIALIZE GRAPH FROM SKELETON");
		}
	    }
	}
    }
	
    final public Graph GetInitializerGraph() {
	return initializerGraph;
    }

    final public void FilterSaltAndPepperNoise() {
	for(int j=0;j<GetYSize();j++) {
	    for(int i=GetXSize() - 1;i>=0;i--) {
		if (GetPixel(i,j,0)) { // black pixel
		    if (NumOfNeighbors(i,j) == 0) {
			SetPixel(i,j,0);
		    }
		}
	    }
	}
	boolean found = true;
	while(found) {
	    found = false;
	    for(int j=0;j<GetYSize();j++) {
		for(int i=GetXSize() - 1;i>=0;i--) {
		    if (!GetPixel(i,j,0)) { // white pixel
			boolean makeItBlack = true;
			boolean foundWhiteCorner = false;
			for(int k=1;k<=8 && makeItBlack;k++) {
			    if (k%2 != 0) {
				if (!neighbor(i,j,k)) { // white 4-neighbor
				    makeItBlack = false;
				}
			    }
			    else {
				if (!neighbor(i,j,k)) {
				    if (!foundWhiteCorner) // first white corner-neighbor
					foundWhiteCorner = true;
				    else // after first white corner-neighbor
					makeItBlack = false;
				}
			    }
			}
			if (makeItBlack) { // && foundWhiteCorner) {
			    found = true;
			    SetPixel(i,j,1);
			}
		    }
		}
	    }
	}	
    }

final class Pixel extends Object {
    int i;
    int j;
    
    Pixel(int in_i,int in_j) {
	i = in_i;
	j = in_j;
    }

    public boolean equals(Object pixel) {
	try {
	    return ((Pixel)pixel).i == i && ((Pixel)pixel).j == j;
	}
	catch (ClassCastException e) {
	    return false;
	}
    } 

    final public String toString() {
	return "(" + i + "," + j + ")";
    }
}

}



