package LinearAlgebra;

import java.io.*;
import java.util.*;
import Debug.*;

final public class Graph extends GraphAbstract implements Loadable {
    Vector edges;

    public Graph() {
	super();
	edges = new Vector(100);
    }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                  Abstract functions of GraphAbstract BEGIN
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    final public int GetVektorIndex1OfEdgeAt(int index) {
	return ((GraphEdge)edges.elementAt(index)).vektorIndex1;
    }

    final public int GetVektorIndex2OfEdgeAt(int index) {
	return ((GraphEdge)edges.elementAt(index)).vektorIndex2;
    }

    final public int GetNumOfLineSegments() {
	return edges.size();
    }
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                  Abstract functions of GraphAbstract END
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    final public void Load(File f,Debug d) throws FileNotFoundException, IOException {
	FileInputStream fin = new FileInputStream(f);
	DataInputStream din = new DataInputStream(fin);
	Load(din,d);
	din.close();
	fin.close(); 
    }

    final public void Load(File f) throws FileNotFoundException, IOException {
	FileInputStream fin = new FileInputStream(f);
	DataInputStream din = new DataInputStream(fin);
	Load(din);
	din.close();
	fin.close(); 
    }

    final public void Load(DataInputStream din,Debug d) throws IOException {
	throw new RuntimeException("NOT YET IMPLEMENTED\n");
    }

    final public void Load(DataInputStream din) throws IOException {
	Reset();
	String line = new String(din.readLine());
	StringTokenizer t = new StringTokenizer(line);
	int dimension = (new Integer(t.nextToken())).intValue();
	line = new String(din.readLine());
	while(din.available() > 0 && !line.equals("") ) {
	    t = new StringTokenizer(line);
	    if (dimension == 2) {
		try {
		    AddPoint(new Vektor2D(t));
		}
		catch (NoSuchElementException e1) {
		}
		catch (NumberFormatException e) {
		}
	    }
	    else
		try {
		    AddPoint(new VektorDD(t));
		}
		catch (NoSuchElementException e1) {
		}
		catch (NumberFormatException e) {
		}
	    line = new String(din.readLine());
	}
	line = new String(din.readLine());
	while(din.available() > 0) {
	    t = new StringTokenizer(line);
	    try {
		AddEdge(new GraphEdge(t));
	    }
	    catch (NoSuchElementException e1) {
	    }
	    catch (NumberFormatException e) {
	    }
	    line = new String(din.readLine());
	}
    }
    
    final public void AddEdge(GraphEdge edge) {
	edges.insertElementAt(edge,GetNumOfLineSegments());
    }
    
    final public void AddEdge(int vektorIndex1,int vektorIndex2) {
	AddEdge(new GraphEdge(vektorIndex1,vektorIndex2));
    }
    
    final public void AddEdge(Vektor vektor1,Vektor vektor2) {
	int vi1 = FindPoint(vektor1);
	int vi2 = FindPoint(vektor2);
	if (vi1 == vi2 && vektor1 != vektor2) {
	    vi1 = Math.min(vi1,vi2);
	    vi2 = points.indexOf(vektor2,vi1 + 1);
	}
	AddEdge(vi1,vi2);
    }

    // Assume that all incident edges are deleted
    final public void DeletePointAt(int index) {
	super.DeletePointAt(index);
	for(int i=0;i<GetNumOfLineSegments();i++) {
	    ((GraphEdge)edges.elementAt(i)).DecrementPointIndexes(index);
	}
    }
	
    final public void DeleteEdgeAt(int index) {
	try {
	    edges.removeElementAt(index);
	}
	catch (ArrayIndexOutOfBoundsException e) {
	    throw new ArrayIndexOutOfBoundsException("Myerror: index = " + index + 
						     " size = " + GetNumOfLineSegments() + "\n");
	}
    }

    final protected int FindEdge(GraphEdge edge) {
	return edges.indexOf(edge);
    }

    final public int FindEdge(Vektor vektor1,Vektor vektor2) {
	GraphEdge edge = new GraphEdge(FindPoint(vektor1),FindPoint(vektor2));
	return FindEdge(edge);
    }

    // Find adjoint vertex, and delete connecting edge
    final public Vektor FindNextVertex(Vektor vektor) {
	int vektorIndex = FindPoint(vektor);
	GraphEdge edge;
	for(int i=0;i<GetNumOfLineSegments();i++) {
	    edge = (GraphEdge)edges.elementAt(i);
	    if (edge.vektorIndex1 == vektorIndex) {
		vektorIndex = edge.vektorIndex2;
		DeleteEdgeAt(i);
		return GetPointAt(vektorIndex);
	    }
	    else if (edge.vektorIndex2 == vektorIndex) {
		vektorIndex = edge.vektorIndex1;
		DeleteEdgeAt(i);
		return GetPointAt(vektorIndex);
	    }
	}
	return null;
    }

final class GraphEdge extends Object {
    int vektorIndex1;
    int vektorIndex2;

    public GraphEdge(int in_vektorIndex1,int in_vektorIndex2) {
	vektorIndex1 = in_vektorIndex1;
	vektorIndex2 = in_vektorIndex2;
    }

    public GraphEdge(StringTokenizer t)  throws NoSuchElementException, NumberFormatException {
	vektorIndex1 = (new Long(t.nextToken())).intValue();
	vektorIndex2 = (new Long(t.nextToken())).intValue();
    }

    final public boolean equals(Object edge) {
	try {
	    return ( (((GraphEdge)edge).vektorIndex1 == vektorIndex1 && 
		      ((GraphEdge)edge).vektorIndex2 == vektorIndex2) ||
		     (((GraphEdge)edge).vektorIndex1 == vektorIndex2 && 
		      ((GraphEdge)edge).vektorIndex2 == vektorIndex1) );
	}
	catch (ClassCastException e) {
	    return false;
	}
    } 

    final public void DecrementPointIndexes(int lowerIndex) {
	if (vektorIndex1 > lowerIndex)
	    vektorIndex1--;
	if (vektorIndex2 > lowerIndex)
	    vektorIndex2--;
    }
}    
}



