package Paintable;

import java.awt.*;
import java.util.*;

import AWT.*;
import LinearAlgebra.Vektor;

final public class ProjectionPlaneSettingDialog extends Dialog {
    Panel panel;
    Button doneButton;
    ProjectionPlane projectionPlane;
    Recallable recallable;

    public ProjectionPlaneSettingDialog(Frame frame,ProjectionPlane in_projectionPlane,Recallable in_recallable) {
	super(frame,"Projection plane",false);
	recallable = in_recallable;
	Constructor(in_projectionPlane);
    }    

    public ProjectionPlaneSettingDialog(Frame frame,ProjectionPlane in_projectionPlane) {
	super(frame,"Projection plane",false);
	Constructor(in_projectionPlane);
    }    

    private void Constructor(ProjectionPlane in_projectionPlane) {
	projectionPlane = in_projectionPlane;
	panel = new Panel();
  	setLayout(new BorderLayout());
  	add("South",panel);
  	doneButton = new Button("Done");
  	add("North",doneButton);
  	SetupPanel();
	validate();
	pack();
    }    

    int dimension;

    Panel settingsPanel;
    Panel originPanel;
    Panel axisXPanel;
    Panel axisYPanel;
    ScrollbarTextFieldPanel [] originPanels;
    ScrollbarTextFieldPanel [] axisXPanels;
    ScrollbarTextFieldPanel [] axisYPanels;
    public static int FACTOR = 100; // resetable by applications
    public static int MAX_XY = 103; 
    public static int MAX_ORIGIN = 1000; 

    Panel controlPanel;
    Button resetButton;
    Button firstTwoPrincipalComponentsButton;
    Checkbox normalizeCheckbox;
    Checkbox orthogonalizeCheckbox;
    Checkbox repaintCheckbox;

    final private void SetupPanel() {
	dimension = projectionPlane.axisX.Dimension();

	settingsPanel = new Panel();
	settingsPanel.setLayout(new GridLayout(1,0));
	axisXPanel = new Panel();
	axisXPanel.setLayout(new GridLayout(0,1));
	axisXPanel.add(new Label("X axis                        "));
	axisYPanel = new Panel();
	axisYPanel.setLayout(new GridLayout(0,1));
	axisYPanel.add(new Label("Y axis                        "));
	originPanel = new Panel();
	originPanel.setLayout(new GridLayout(0,1));
	originPanel.add(new Label("Origin                        "));

	axisXPanels = new ScrollbarTextFieldPanel[dimension];
	axisYPanels = new ScrollbarTextFieldPanel[dimension];
	originPanels = new ScrollbarTextFieldPanel[dimension];
	String space;
	if (dimension <= 10)
	    space = "  ";
	else if (dimension <= 100)
	    space = "   ";
	else 
	    space = "    ";
	for(int i=0;i<dimension;i++) {
	    axisXPanels[i] = new ScrollbarTextFieldPanel(space + i + ":",
							 (int)(projectionPlane.axisX.GetCoords(i) * FACTOR),
							 -MAX_XY + 3,MAX_XY,true,FACTOR);
	    axisXPanel.add(axisXPanels[i]);
	    axisYPanels[i] = new ScrollbarTextFieldPanel(space + i + ":",
							 (int)(projectionPlane.axisY.GetCoords(i) * FACTOR),
							 -MAX_XY + 3,MAX_XY,true,FACTOR);
	    axisYPanel.add(axisYPanels[i]);
	    originPanels[i] = new ScrollbarTextFieldPanel(space + i + ":",
							 (int)(projectionPlane.origin.GetCoords(i) * FACTOR),
							 -MAX_ORIGIN,MAX_ORIGIN,true,FACTOR);
	    originPanel.add(originPanels[i]);
	}
	
	settingsPanel.add(axisXPanel);
	settingsPanel.add(axisYPanel);
	settingsPanel.add(originPanel);
	panel.add(settingsPanel);

	
	controlPanel = new Panel();
	controlPanel.setLayout(new GridLayout(1,0));

      	resetButton = new Button("Reset");
	firstTwoPrincipalComponentsButton = new Button("First 2 PC's");
	normalizeCheckbox = new Checkbox("Normalize",false);
	orthogonalizeCheckbox = new Checkbox("Orthogonalize",false);
	repaintCheckbox = new Checkbox("Repaint",true);
	
	controlPanel.add(resetButton);	
	controlPanel.add(firstTwoPrincipalComponentsButton);	
	controlPanel.add(normalizeCheckbox);	
	controlPanel.add(orthogonalizeCheckbox);	
	controlPanel.add(repaintCheckbox);	
	add("Center",controlPanel);
    }

    final private void ConditionalRepaint() {
	if (repaintCheckbox.getState()) {
	    try {
		recallable.Recall();
	    }
	    catch(NullPointerException e) {}
	}
    }

    final public boolean action(Event event, Object arg) {
	// Done
	if (event.target == doneButton) {
	    postEvent(new Event(this,Event.WINDOW_DESTROY,""));
	}
	else if (event.target == resetButton) {
	    projectionPlane.Reset(dimension);
	    UpdateScrollbars();
	    ConditionalRepaint();
	}
	else if (event.target == firstTwoPrincipalComponentsButton) {
	    projectionPlane.SetToFirstTwoPrincipalComponents();
	    UpdateScrollbars();
	    ConditionalRepaint();
	}
	else if (event.target == normalizeCheckbox) {
	    if (normalizeCheckbox.getState()) {
		projectionPlane.Normalize();
		UpdateScrollbars();
		ConditionalRepaint();
	    }
	}
	else if (event.target == orthogonalizeCheckbox) {
	    if (orthogonalizeCheckbox.getState()) {
		projectionPlane.Orthogonalize(0); // X axis is unchanged
		UpdateScrollbars();
		ConditionalRepaint();
	    }
	}
	else if (event.target == repaintCheckbox) {
	    ConditionalRepaint();
	}
	return true;
    }

    boolean realEvent = true;
    int countNonrealEvents;

    final public boolean handleEvent(Event event) {
	// Done
	if (event.id == Event.WINDOW_DESTROY) {
	    dispose();
	}
	else {
	    for(int i=0;i<dimension;i++) {
		if (event.target == axisXPanels[i] && ScrollbarTextFieldPanel.ScrollbarEvent(event)) {
		    projectionPlane.axisX.SetCoord(i,(((Integer)event.arg).intValue())/(double)FACTOR);
		    if (!realEvent) {
			if (countNonrealEvents == 3 * dimension)
			    realEvent = true;
			else
			    countNonrealEvents++;
		    }
		    if (realEvent) {
			if (orthogonalizeCheckbox.getState()) {
			    projectionPlane.Orthogonalize(0); // X axis is unchanged
			}
			if (normalizeCheckbox.getState()) {
			    projectionPlane.Normalize();
			}
			if (orthogonalizeCheckbox.getState() || normalizeCheckbox.getState())
			    UpdateScrollbars();
			if (repaintCheckbox.getState()) {
			    try {
				recallable.Recall();
			    }
			    catch(NullPointerException e) {}
			}
		    }		    
		    i = dimension;
		}
		else if (event.target == axisYPanels[i] && ScrollbarTextFieldPanel.ScrollbarEvent(event)) {
		    projectionPlane.axisY.SetCoord(i,(((Integer)event.arg).intValue())/(double)FACTOR);
		    if (!realEvent) {
			if (countNonrealEvents == 3 * dimension)
			    realEvent = true;
			else
			    countNonrealEvents++;
		    }
		    if (realEvent) {
			if (orthogonalizeCheckbox.getState()) {
			    projectionPlane.Orthogonalize(1); // Y axis is unchanged
			}
			if (normalizeCheckbox.getState()) {
			    projectionPlane.Normalize();
			}
			if (orthogonalizeCheckbox.getState() || normalizeCheckbox.getState())
			    UpdateScrollbars();
			if (repaintCheckbox.getState()) {
			    try {
				recallable.Recall();
			    }
			    catch(NullPointerException e) {}
			}
		    }
		    i = dimension;
		}
		else if (event.target == originPanels[i] && ScrollbarTextFieldPanel.ScrollbarEvent(event)) {
		    projectionPlane.origin.SetCoord(i,(((Integer)event.arg).intValue())/(double)FACTOR);
		    if (!realEvent) {
			if (countNonrealEvents == 3 * dimension)
			    realEvent = true;
			else
			    countNonrealEvents++;
		    }
		    if (realEvent && repaintCheckbox.getState()) {
			try {
			    recallable.Recall();
			}
			catch(NullPointerException e) {}
		    }
		    i = dimension;
		}
	    }
	}
        return super.handleEvent(event);    
    }

    final public void UpdateScrollbars() {
	realEvent = false;
	countNonrealEvents = 0;
	for(int i=0;i<dimension;i++) {
	    originPanels[i].setValue((int)(projectionPlane.origin.GetCoords(i) * FACTOR));
	    axisXPanels[i].setValue((int)(projectionPlane.axisX.GetCoords(i) * FACTOR));
	    axisYPanels[i].setValue((int)(projectionPlane.axisY.GetCoords(i) * FACTOR));    
	}
    }
}



