package PrincipalCurve;

import java.io.*;
import java.util.*;

import Bitmap.*;
import Debug.*;
import LinearAlgebra.*;

public class  PrincipalCurveAlgorithmForSkeletonization extends PrincipalCurveAlgorithm {
    Template template;
    Template savedTemplate;
    // Debug strings
    public final static String DEBUG_RESTRUCTURE = "Restructuring... ";
    public final static String DEBUG_SKELETONIZE = "Skeletonizing... ";
    public final static String DEBUG_INIT = "Initizalizing to skeleton graph... ";
    public final static String DEBUG_FIT_AND_SMOOTH = "Fitting and smoothing... ";
    public final static String DEBUG_ALL = "Optimizing skeleton... ";

    public PrincipalCurveAlgorithmForSkeletonization(Template in_template,PrincipalCurveClass initialCurve) {
	super(initialCurve);
	savedTemplate = template = (Template)in_template.Clone();
    }

    final public synchronized void Skeletonize() {
	Debug fd = new Debug(DEBUG_SKELETONIZE);
	try {
	    wait(0,1);
	} catch (InterruptedException e) {}

	template = (Template)savedTemplate.Clone();
	template.Skeletonize();
	
	fd.Terminate();
	try {
	    wait(0,1);
	} catch (InterruptedException e) {}
    }	   
   
    final public synchronized void Initialize() {
	//	super.Initialize();
	Debug fd = new Debug(DEBUG_INIT);
	try {
	    wait(0,1);
	} catch (InterruptedException e) {}

	double rate = 1;
	Template localTemplate = (Template)template.Clone();
	if (GlobalsSkeletonization.changeResolution) {
	    rate = (double)localTemplate.GetXSize()/GlobalsSkeletonization.changeResolutionX;
	    localTemplate.Normalize(GlobalsSkeletonization.changeResolutionX,
			       (int)((double)GlobalsSkeletonization.changeResolutionX*localTemplate.GetYSize()/
				     localTemplate.GetXSize() + 0.5));
	}
	localTemplate.SetInitializerGraph();
	Graph initalizerGraph = localTemplate.GetInitializerGraph();
	if (GlobalsSkeletonization.changeResolution) {
	    for(int i=0;i<initalizerGraph.GetSize();i++)
		initalizerGraph.GetPointAt(i).MulEqual(rate);
	}
	((PrincipalCurveOfTemplate)principalCurve).InitializeToGraph(initalizerGraph);

	fd.Terminate();
	try {
	    wait(0,1);
	} catch (InterruptedException e) {}	
    }

    final public synchronized void Continue() {
//      super.Continue();

	Debug fd = new Debug(DEBUG_ALL);
	try {
	    wait(0,1);
	} catch (InterruptedException e) {}

	OuterStep();
	Restructure();
	OuterStep();

	fd.Terminate();
	try {
	    wait(0,1);
	} catch (InterruptedException e) {}
	
    }	   
    
    final public synchronized void OuterStep() {
	Debug fd = new Debug(DEBUG_FIT_AND_SMOOTH);	
	try {
	    wait(0,1);
	} catch (InterruptedException e) {}
	iteration = 0;
	boolean cont = true;
	while( !stop && (iteration < 2  || cont) && iteration < 500){
	    iteration++;
	    fd.Iterate();
	    try {
		wait(0,1);
	    } catch (InterruptedException e) {}
	    cont = InnerStep();
	}
	fd.Terminate();
	try {
	    wait(0,1);
	} catch (InterruptedException e) {}
    }	      

    final public synchronized void Restructure() {
	Debug fd = new Debug(DEBUG_RESTRUCTURE);
	try {
	    wait(0,1);
	} catch (InterruptedException e) {}

	((PrincipalCurveOfTemplate)principalCurve).Restructure();

	fd.Terminate();	  
    }

}
