package PrincipalCurve;

import java.awt.*;
import java.util.*;
import LinearAlgebra.*;
import Paintable.*;
import PrincipalCurve.Vertex.*;
import PrincipalCurve.Vertex.Elements.*;

final public class PrincipalCurveOfTemplate extends PrincipalCurveClass {
    Paths paths;
	
    public PrincipalCurveOfTemplate(Sample in_sample) {
	super(in_sample);
	Globals.relativeChangeInCriterionThreshold = 0.004;
	Globals.penaltyCoefficient = 0.1;
	paths = null;
    }
   

    final protected void JoinVertices(Curve curve,int k,int j) {
	// EndVertex + EndVertex = LineVertex
	if (GetVertexAt(k) instanceof EndVertex && curve.GetPointAt(j) instanceof EndVertex)
	    SetPointAt(new LineVertex((EndVertex)GetVertexAt(k),
				      (EndVertex)curve.GetPointAt(j)),k);
	//  LineVertex + EndVertex = StarOfThreeVertex
	else if (GetVertexAt(k) instanceof LineVertex && curve.GetPointAt(j) instanceof EndVertex)
	    SetPointAt(new StarOfThreeVertex((LineVertex)GetVertexAt(k),
					     (EndVertex)curve.GetPointAt(j)),k);
	// EndVertex + LineVertex = StarOfThreeVertex
	else if (GetVertexAt(k) instanceof EndVertex && curve.GetPointAt(j) instanceof LineVertex)
	    SetPointAt(new StarOfThreeVertex((LineVertex)curve.GetPointAt(j),
					     (EndVertex)GetVertexAt(k)),k);
	// LineVertex + LineVertex = StarOfFourVertex
	else if (GetVertexAt(k) instanceof LineVertex && curve.GetPointAt(j) instanceof LineVertex)
	    SetPointAt(new StarOfFourVertex((LineVertex)GetVertexAt(k),
					    (LineVertex)curve.GetPointAt(j)),k);
	// StarOfThreeVertex + EndVertex = StarOfFourVertex = XVertex
	else if (GetVertexAt(k) instanceof StarOfThreeVertex && curve.GetPointAt(j) instanceof EndVertex)
	    SetPointAt(new StarOfFourVertex((StarOfThreeVertex)GetVertexAt(k),
					    (EndVertex)curve.GetPointAt(j)),k);
	// EndVertex + StarOfThreeVertex = StarOfFourVertex = XVertex
	else if (GetVertexAt(k) instanceof EndVertex && curve.GetPointAt(j) instanceof StarOfFourVertex)
	    SetPointAt(new StarOfFourVertex((StarOfThreeVertex)curve.GetPointAt(j),
					    (EndVertex)GetVertexAt(k)),k);
 	else
 	    throw new RuntimeException("NOT KNOWN INITIAL JOIN TYPE!!\n" + 
 				       curve.GetPointAt(j) + "\n" +  GetVertexAt(k) + "\n");
    }
    
    // Maintaining neighbors, so if angle is not penalized at a vertex, it is not penalized
    // either when a neighbor vertex is relocated.
    protected void MaintainNeighbors() {
 	for(int i=0;i<GetSize();i++) {
 	    if (GetVertexAt(i) instanceof YVertex || 
		GetVertexAt(i) instanceof StarOfThreeVertex ||
		GetVertexAt(i) instanceof StarOfFourVertex)
		GetVertexAt(i).MaintainNeighbors();
	    if (GetVertexAt(i) instanceof StarOfFourVertex) {
		SetPointAt(GetVertexAt(i).Restructure(),i);  // making it an X-vertex
	    }		
	}
    }

    // Assume that the label of each vertex is its degree, and that at least one of any two
    // neighbors has degree <= 2 (no neighbor joinvertices).
    final public void InitializeToGraph(Graph graph) {
	Reset();	
	VektorDDLabeled vertex,newVertex,multiVertex;
	double [] coords = {2*sample.GetRadius()*Utilities.MyMath.epsilon,2*sample.GetRadius()*Utilities.MyMath.epsilon};
	VektorDDLabeled verySmallVektor = new VektorDDLabeled(coords,0);

	// filtering vertices with degree > 4
	for(int i=0;i<graph.GetSize();i++) {
	    vertex = (VektorDDLabeled)graph.GetPointAt(i);
	    if (vertex.GetLabel() > 4) {
		multiVertex = vertex;
		int edgeIndex;
		// Putting all neighbors in a sample, deleting all incident edges
		Sample neighbors = new Sample();
		for(int j=0;j<graph.GetSize();j++) {
		    edgeIndex = graph.FindEdge(vertex,graph.GetPointAt(j));
		    if (edgeIndex > -1) {
			neighbors.AddPoint(graph.GetPointAt(j));
			graph.DeleteEdgeAt(edgeIndex);
		    }
		}
		// Deleting the original point, resetting loop counter
		graph.DeletePointAt(i);
		i--;
		// Cluster center will be the center of mass of neighbors. Neighbors are 
		// traced in order of closeness to the cluster center.
		newVertex = (VektorDDLabeled)neighbors.GetPointAt(0).Clone();
		newVertex.AddEqual(verySmallVektor); // for differentiating from the existing vertex
		graph.AddPoint(newVertex);
		double minDist;
		int minIndex;
		int clusterSize = 0;
		boolean first = true;
		while (neighbors.GetSize() > 0) {
		    minDist = newVertex.Dist2(neighbors.GetPointAt(0));
		    minIndex = 0;
		    for(int j=1;j<neighbors.GetSize();j++) {
			if (minDist > newVertex.Dist2(neighbors.GetPointAt(j))) {
			    minDist = newVertex.Dist2(neighbors.GetPointAt(j));
			    minIndex = j;
			}
		    }
		    // Until 3 neighbors, the center is reset to the new center of mass. When the 4th 
		    // neighbor is processed, a new cluster center is created, and the two cluster centers
		    // are connected by an edge.
		    if ((first && clusterSize < 3) || clusterSize < 2) {
			clusterSize++;
			newVertex.MulEqual((clusterSize - 1.0)/clusterSize);
			newVertex.AddEqual(neighbors.GetPointAt(minIndex).Div(clusterSize));
			newVertex.AddEqual(verySmallVektor); // for differentiating from existing vertices
			newVertex.SetLabel(clusterSize + 1);
			graph.AddEdge(newVertex,neighbors.GetPointAt(minIndex));
			neighbors.DeletePointAt(minIndex);
		    }
		    else { // clustersize == 3
			first = false;
			clusterSize = 0;
			newVertex.SetLabel(4);
			vertex = newVertex;

			newVertex = (VektorDDLabeled)neighbors.GetPointAt(minIndex).Clone();			
			newVertex.AddEqual(verySmallVektor); // for differentiating from the existing vertex
			graph.AddPoint(newVertex);
			graph.AddEdge(vertex,newVertex);
			
			// For bringing them closer to each other
			vertex.AddEqual(multiVertex.Mul(1.0/4));
			vertex.DivEqual(5.0/4);
		    }
		}			
		// For bringing them closer to each other
  		newVertex.AddEqual(multiVertex.Mul(1.0/newVertex.GetLabel()));
  		newVertex.DivEqual((newVertex.GetLabel() + 1.0)/newVertex.GetLabel());
	    }
	}
	boolean foundVertex = true;
	SetOfCurves setOfCurves = new SetOfCurves();
	int i;
	while(foundVertex) {
	    foundVertex = false;
	    i = -1;
	    while(!foundVertex && i < graph.GetSize() - 2) {
		i++;
		if (((VektorDDLabeled)graph.GetPointAt(i)).GetLabel() > 0)
		    foundVertex = true;
	    }
	    if (foundVertex) {
		setOfCurves.StartNewCurve();
		vertex = (VektorDDLabeled)graph.GetPointAt(i);
		setOfCurves.AddPoint(vertex);
		while(vertex.GetLabel() > 0) {
		    vertex.SetLabel(vertex.GetLabel() - 1);
		    newVertex = vertex; // for debugging
		    // find adjoint vertex, and delete connecting edge
		    vertex = (VektorDDLabeled)graph.FindNextVertex(vertex); 
		    setOfCurves.AddPoint(vertex);
		    try {
			vertex.SetLabel(vertex.GetLabel() - 1);
		    }
		    catch (NullPointerException e) {
			String s = graph.toString();
			s += "\nvertex: " + vertex.toString();
			throw new RuntimeException("Initialization error/n" + s);
		    }
		}
	    }
	}
	InitializeToCurves(setOfCurves,0);
    }

    int s = 1;
    double thickness;

    final public void Restructure() {
	paths = new Paths(this);
	RepartitionVoronoiRegions();
	thickness = 0;
	PrincipalCurveSampleVektor samplePoint;
	for(int i=0;i<sample.GetSize();i++) {	
	    samplePoint = GetSamplePointAt(i);
	    thickness += samplePoint.GetWeight() * samplePoint.Dist2(GetEdgeAt(samplePoint.GetIndexOfNearestSegment()));
	}
	thickness *= 4.0/sample.GetSize();
 	System.out.println("Restructure");	
 	System.out.println("thickness = " + thickness);
	paths = new Paths(this);
	System.out.println(paths);

      	System.out.println("Deleting short branches");
  	DeleteOneEdgeBranches();
 	DeleteShortBranches(GlobalsSkeletonization.tauBranch*thickness);
	System.out.println(paths);

       	System.out.println("Merging StarOfThreeVertices");
  	MergeStarOfThreeVertices(GlobalsSkeletonization.epsStar3);
 	System.out.println(paths);

      	System.out.println("Deleting small holes");
 	EliminateSmallHoles(GlobalsSkeletonization.tauLoop*thickness);
	System.out.println(paths);

      	System.out.println("Updating vertices");
 	UpdateStarVertices();
	System.out.println(paths);

	System.out.println("thickness = " + thickness);
	FilterVerticesImaiIri(GlobalsSkeletonization.tauFilter,false,true);	
    }
    
    final private void AddVertices(double lengthThreshold) {
	boolean added = true;
	while(added) {
	    added = false;
	    for(int i=0;i<GetNumOfLineSegments();i++) {
		if (GetEdgeAt(i).GetLength() > lengthThreshold) {
		    InsertMidPoint(i);
		    added = true;
		}
	    }
	}
	paths = new Paths(this);	
    }
	
    final public void FilterVerticesImaiIri(double epsilon, boolean lineVertexHasBeenKept,boolean del) {
	paths = new Paths(this);	
	for(int i=0;i<GetSize();i++)
	    GetVertexAt(i).label = 0;
	for(int i=0;i<paths.GetSize();i++) {
	    Vertex previousVertex = paths.GetPathAt(i).firstVertex;
	    Vertex vertex = paths.GetPathAt(i).secondVertex;
	    Vertex nextVertex;
	    double angle = 0;
	    double length = previousVertex.Dist2(vertex);
	    boolean lineVertexHasBeenKeptLocal = lineVertexHasBeenKept;
	    Vertex keptVertex = paths.GetPathAt(i).firstVertex;
	    Vector vertices = new Vector(paths.GetSize());
	    vertices.addElement(vertex);
	    while(!(vertex.equals(paths.GetPathAt(i).lastVertex))) {
		nextVertex = ((LineVertex)vertex).GetOppositeVertex(previousVertex);
		length += vertex.Dist2(nextVertex);
		LineSegment lineSegment = new LineSegmentObject(keptVertex,nextVertex);
		boolean cont = true;
		for(int j=0;j<vertices.size() && cont;j++) {
		    if (((VektorDD)(vertices.elementAt(j))).Dist2(lineSegment) > epsilon)
			cont = false;
		}
		if (cont) {
		    if (!lineVertexHasBeenKeptLocal && length > paths.GetPathAt(i).length/2) {
			lineVertexHasBeenKeptLocal = true;
			keptVertex = vertex;
			vertices.removeAllElements();
		    }
		    else {
			vertex.label = 1;
		    }
		}
		else {
		    lineVertexHasBeenKeptLocal = true;
		    keptVertex = vertex;
		    vertices.removeAllElements();
		}
		previousVertex = vertex;
		vertex = nextVertex;
		vertices.addElement(vertex);
	    }
	}
	if (del) {
	    for(int i=0;i<GetSize();i++) {
		if (GetVertexAt(i).label == 1) {
		    DeleteLineVertexAt(i);
		    i--;
		}
	    }
	    paths = new Paths(this);
	}
    }
    
    // Deleting all EndVertices that are joined directly to a join vertex. It is for noise reduction, and
    // it is also necessary to avoid degrading EndVertices in MergeStarOfThreeVertices().
    final private void DeleteOneEdgeBranches() {
	for(int i=0;i<paths.GetSize();i++) {
	    if (paths.GetPathAt(i).OneEdgeBranch()) {
		DeletePath(i);
		i--;
	    }
	}
	paths = new Paths(this);
    }

    final private void EliminateSmallHoles(double threshold) {
	boolean deleted = true;
	// Loops of one path
	while(deleted) {
	    deleted = false;
	    for(int i=0;i<paths.GetSize() && !deleted;i++) {
		if (paths.GetPathAt(i).length < threshold && paths.GetPathAt(i).Loop()) {
		    DeletePath(i);
		    deleted = true;
		    paths = new Paths(this);
		}
	    }
	}
	deleted = true;
	// Loops of two paths (delete the longest path)
	while(deleted) {
	    deleted = false;
	    for(int i=0;i<paths.GetSize() && !deleted;i++) {
		for(int j=i+1;j<paths.GetSize() && !deleted;j++) {
		    if (paths.GetPathAt(i).length + paths.GetPathAt(j).length < 2*threshold &&
			Hole(paths.GetPathAt(i),paths.GetPathAt(j))) {
			if (paths.GetPathAt(i).length > paths.GetPathAt(j).length) {
			    DeletePath(i);
			}
			else {
			    DeletePath(j);
			}
			deleted = true;
			paths = new Paths(this);
		    }
		}
	    }
	}
	deleted = true;
	// Loops of three paths (delete the longest path)
	while(deleted) {
	    deleted = false;
	    for(int i=0;i<paths.GetSize();i++) {
		for(int j=i+1;j<paths.GetSize();j++) {
		    for(int k=j+1;k<paths.GetSize();k++) {
			if (paths.GetPathAt(i).length + paths.GetPathAt(j).length + paths.GetPathAt(k).length < 
			    3*threshold && Hole(paths.GetPathAt(i),paths.GetPathAt(j),paths.GetPathAt(k))) {
			    if (paths.GetPathAt(i).length > paths.GetPathAt(j).length && 
				paths.GetPathAt(i).length > paths.GetPathAt(k).length) {
				DeletePath(i);
			    }
			    else if (paths.GetPathAt(j).length > paths.GetPathAt(i).length && 
				 paths.GetPathAt(j).length > paths.GetPathAt(k).length){
				DeletePath(j);
			    }
			    else {
				DeletePath(k);
			    }
			    deleted = true;
			    paths = new Paths(this);
			}
		    }
		}
	    }		    
	}
    }

    final private double LLessThanSixty(double alpha) {
	return (1/Math.sin(alpha/2) - 1)/(Math.sin(alpha/2) + 1);
    }

    final private double LMoreThanSixty(double alpha) {
	return Math.cos(alpha)/(Math.sin(alpha)*Math.cos(alpha/2));
    }

    final private double Alpha(double gamma) {    
	return gamma/2;
//	return 2*gamma - Math.PI;
    }

    final private double AlphaDegree(double gamma) {    
	return gamma/2;
//	return 2*gamma - 180;
    }

    final private double EstimatedLengthForMerge(Path path) {
	Vertex [] neighbors  = path.firstVertex.GetNeighbors();
	Vertex [] otherNeighbors = new Vertex[2];
	int l1 = 0;
	for(int l=0;l<3;l++)
	    if (!neighbors[l].equals(path.secondVertex))
		otherNeighbors[l1++] = neighbors[l];
	double cosAngle1 = path.firstVertex.CosAngle(otherNeighbors[0],otherNeighbors[1]);

	neighbors = path.lastVertex.GetNeighbors();
	otherNeighbors = new Vertex[2];
	l1 = 0;
	for(int l=0;l<3;l++)
	    if (!neighbors[l].equals(path.beforeVertex))
		otherNeighbors[l1++] = neighbors[l];
	double cosAngle2 = path.lastVertex.CosAngle(otherNeighbors[0],otherNeighbors[1]);

	double angle1 = Math.acos(cosAngle1);
	double angle1Degree = 180*angle1/Math.PI;
	double length1;
 	System.out.println("gamma1 = " + angle1Degree);
 	System.out.println("alpha1 = " + AlphaDegree(angle1Degree));
	if (angle1Degree/2 > 60)
	    length1 = 0.5*thickness*LMoreThanSixty(Alpha(angle1));
	else
	    length1 = 0.5*thickness*LLessThanSixty(Alpha(angle1));
	System.out.println("length1 = " + length1);
	
	double angle2 = Math.acos(cosAngle2);
	double angle2Degree = 180*angle2/Math.PI;
	double length2;
 	System.out.println("gamma2 = " + angle2Degree);
 	System.out.println("alpha2 = " + AlphaDegree(angle2Degree));
	if (angle2Degree/2 > 60)
	    length2 = 0.5*thickness*LMoreThanSixty(Alpha(angle2));
	else
	    length2 = 0.5*thickness*LLessThanSixty(Alpha(angle2));
	System.out.println("length2 = " + length2);
	
	System.out.println();
	System.out.println("length1 + length2 = " + (length1 + length2));
 	System.out.println(path);
	return length1 + length2;
    }

    final private void MergeStarOfThreeVertices(double factor) {
 	System.out.println("factor = " + factor);
	boolean merged = true;
	while(merged) {
	    merged = false;
	    double minProportion = Double.POSITIVE_INFINITY,prop;
	    int minIndex = -1;
	    for(int i=0;i<paths.GetSize();i++) {
		Path path = paths.GetPathAt(i);
		if (path.firstVertex.GetDegree() == 3 && path.lastVertex.GetDegree() == 3 
		    && !path.Loop() && (prop = path.length/EstimatedLengthForMerge(path)) < factor && 
		    prop < minProportion) {
		    minProportion = prop;
		    minIndex = i;
		}
	    }
	    if (minIndex != -1) {
		Path path = paths.GetPathAt(minIndex);		
		System.out.println("merge:");
		System.out.println(path);
		System.out.println(minProportion);
		merged = true;
		Sample neighbors = new Sample();
		Vertex [] neighbors1 = path.firstVertex.GetNeighbors();
		Vertex [] neighbors2 = path.lastVertex.GetNeighbors();
		for(int l=0;l<3;l++)
		    if (!neighbors1[l].equals(path.secondVertex))
			neighbors.AddPoint(neighbors1[l]);
		for(int l=0;l<3;l++)
		    if (!neighbors2[l].equals(path.beforeVertex))
			neighbors.AddPoint(neighbors2[l]);
		if (!(neighbors.GetPointAt(0) instanceof LineVertex) || 
		    !(neighbors.GetPointAt(1) instanceof LineVertex) ||
		    !(neighbors.GetPointAt(2) instanceof LineVertex) || 
		    !(neighbors.GetPointAt(3) instanceof LineVertex) )
		    throw new ArithmeticException("MYERROR: Can't merge StarOfThreeVertices");
		DeletePath(minIndex);
		DeletePointAt(FindPoint(path.firstVertex));
		DeletePointAt(FindPoint(path.lastVertex));
		DummyVertex midVertex = new DummyVertex(path.lastVertex.Add(path.firstVertex).Div(2));
		AddPoint(midVertex);
		int [] indices = new int[4];
		for(int l=0;l<4;l++) {
		    indices[l] =  FindPoint(neighbors.GetPointAt(l));
		    AddEdge(new Edge(this,GetSize()-1,indices[l]));
		    }
		int e = GetNumOfLineSegments();
		StarOfFourVertex newVertex = new StarOfFourVertex(midVertex,e-4,e-3,e-2,e-1,
								  (EndVertex)GetVertexAt(indices[0]),
								  (EndVertex)GetVertexAt(indices[1]),
								  (EndVertex)GetVertexAt(indices[2]),
								  (EndVertex)GetVertexAt(indices[3]));
		SetPointAt(newVertex,GetSize()-1);
		for(int l=0;l<4;l++)
		    SetPointAt(new LineVertex((EndVertex)GetVertexAt(indices[l]),(HasOneEdge)newVertex),
			       indices[l]);
		paths = new Paths(this);
	    }
	}
    }
	
	

    // Updating StarOfThreeVertices to either TVertices or YVertices, StarOfFourVertices to XVertices
    final public void UpdateStarVertices() {
	for(int i=0;i<GetSize();i++) {
	    Vektor vertex = GetVertexAt(i);
	    SetPointAt(GetVertexAt(i).Restructure(),i);
	    if (vertex.getClass() != GetVertexAt(i).getClass())
		System.out.println("Updating " + vertex.getClass() + " to " + GetVertexAt(i).getClass());	
	}
	paths = new Paths(this);
    }

    final private double CorrectedLength(Path path,double threshold) {
	if (!path.Branch())
	    return Double.POSITIVE_INFINITY;
	Vertex joinVertex,neighborVertex,lastVertex;
	if (path.firstVertex instanceof EndVertex) {
	    joinVertex = path.lastVertex;
	    neighborVertex = path.beforeVertex;
	    lastVertex = path.GetPointAt(3*threshold,false);
	}
	else {
	    joinVertex = path.firstVertex;			
	    neighborVertex = path.secondVertex;
	    lastVertex = path.GetPointAt(3*threshold,true);
	}
	Vertex [] neighbors = joinVertex.GetNeighbors();
	// finding the smallest angle
	double cosAngle = -1;
	for(int l=0;l<neighbors.length;l++) {
	    if (!neighbors[l].equals(neighborVertex) && 
		joinVertex.CosAngle(lastVertex,neighbors[l]) > cosAngle) {
		cosAngle = joinVertex.CosAngle(lastVertex,neighbors[l]);
	    }
	}
	System.out.println(path.length + " " + path.length * Math.sqrt(1 - cosAngle*cosAngle));	    
	return path.length * Math.sqrt(1 - cosAngle*cosAngle);
    }

    final private boolean ShortBranch(Path path,double correctedLength,double threshold) {
	return path.length < 3*threshold && correctedLength < threshold;
    }

    final private void DeleteShortBranches(double threshold) {
	System.out.println("threshold = " + threshold);
	boolean deleted = true;
	while(deleted) {
	    deleted = false;
	    // finding the shortest short branch
	    int minIndex = -1;
	    double minLength = Double.POSITIVE_INFINITY;
	    for(int i=0;i<paths.GetSize() && !deleted;i++) {
		Path path = paths.GetPathAt(i);
		double correctedLength = CorrectedLength(path,threshold);
		if (ShortBranch(path,correctedLength,threshold) &&
		    correctedLength < minLength) {
		    minIndex = i;
		    minLength = correctedLength;
		}
	    }
	    // deleting if found
	    if (minIndex != -1) {
		System.out.println("Delete: " + paths.GetPathAt(minIndex));
		deleted = true;
		DeletePath(minIndex);
		paths = new Paths(this);		    
	    }
	}
    }

    // Delete the whole path (all LineVertices) from firstVertex in the direction of nextVertex.
    // If the path consists only one edge, only the edge is deleted.
    // First and last vertices are deleted automatically by DeletePointAt() if they are EndVertices. 
    final private void DeletePath(int index) {
 	Vertex firstVertex = paths.GetPathAt(index).firstVertex;
 	Vertex nextVertex = paths.GetPathAt(index).secondVertex;
	paths.DeletePathAt(index);
	if (nextVertex instanceof LineVertex) {
	    Vertex vertex,tempVertex;
	    vertex = nextVertex;
	    nextVertex = ((LineVertex)nextVertex).GetOppositeVertex(firstVertex);
	    while(nextVertex instanceof LineVertex) {
		tempVertex = vertex;
		index = FindPoint(vertex);
		vertex = nextVertex;
		nextVertex = ((LineVertex)nextVertex).GetOppositeVertex(tempVertex);
		DeletePointAt(index);
	    }
	    index = FindPoint(vertex);
	    DeletePointAt(index);
	}
	else {
	    if (firstVertex instanceof EndVertex) 
		DeletePointAt(FindPoint(firstVertex));
	    else if (nextVertex instanceof EndVertex) 
                DeletePointAt(FindPoint(nextVertex));
	    else { // Deleting only the edge
		int vertexIndex1 = FindPoint(firstVertex);
		int vertexIndex2 = FindPoint(nextVertex);
		int edgeIndex = ((HasOneEdge)firstVertex).GetEdgeIndex(nextVertex).edgeIndex;
		// Degrading neighbors
		SetPointAt(GetVertexAt(vertexIndex1).Degrade(nextVertex),vertexIndex1);
		SetPointAt(GetVertexAt(vertexIndex2).Degrade(firstVertex),vertexIndex2);
		// Resetting indices of nearest segment to -1, 
		// so distances are recalculated in RepartitionVoronoiRegions()
		for(int i=0;i<GetEdgeAt(edgeIndex).GetSetSize();i++)
		    GetEdgeAt(edgeIndex).GetSetPointAt(i).SetIndexOfNearestSegment(-1);
		// Deleting edge
		DeleteEdgeAt(edgeIndex);		
	    }
	}
    }

final class Path extends Object {
    Vertex firstVertex;
    Vertex secondVertex;
    Vertex beforeVertex;
    Vertex lastVertex;
    double length;
    int numOfEdges;

    public Path(Vertex in_firstVertex,Vertex in_secondVertex) {
	if (in_firstVertex instanceof LineVertex)
	    throw new ArithmeticException("firstVertex must be  non-LineVertex");
	firstVertex = in_firstVertex;
	secondVertex = in_secondVertex;
	Vertex vertex = firstVertex;
	Vertex nextVertex = secondVertex;
	Vertex tempVertex;
	length = vertex.Dist2(nextVertex);
	numOfEdges = 0;
	while (nextVertex instanceof LineVertex) {
	    vertex.label = 1;
	    tempVertex = vertex;
	    vertex = nextVertex;
	    nextVertex = ((LineVertex)nextVertex).GetOppositeVertex(tempVertex);
	    length += vertex.Dist2(nextVertex);
	    numOfEdges++;
	}
	beforeVertex = vertex;
	beforeVertex.label = 1;
	lastVertex = nextVertex;
	lastVertex.label = 1;
    }

    // for complete circles
    public Path(Vertex in_firstVertex) {
	if (!(in_firstVertex instanceof LineVertex))
	    throw new ArithmeticException("firstVertex must be LineVertex");	
	firstVertex = in_firstVertex;
	secondVertex = firstVertex.GetNeighbors()[0];
	Vertex vertex = firstVertex;
	Vertex nextVertex = secondVertex;
	Vertex tempVertex;
	length = vertex.Dist2(nextVertex);
	numOfEdges = 0;
	while (!(nextVertex.equals(firstVertex))) {
	    vertex.label = 1;
	    tempVertex = vertex;
	    vertex = nextVertex;
	    nextVertex = ((LineVertex)nextVertex).GetOppositeVertex(tempVertex);
	    length += vertex.Dist2(nextVertex);
	    numOfEdges++;
	}
	beforeVertex = vertex;
	beforeVertex.label = 1;
	lastVertex = nextVertex;
	lastVertex.label = 1;
    }

    final public boolean equals(Object object) {
	try {
	    Path path = (Path)object;
	    return ( (path.firstVertex == firstVertex && path.secondVertex == secondVertex && 
		      path.beforeVertex == beforeVertex && path.lastVertex == lastVertex) ||
		     (path.firstVertex == lastVertex && path.secondVertex == beforeVertex && 
		      path.beforeVertex == secondVertex && path.lastVertex == firstVertex) );
	}
	catch(ClassCastException e) {
	    return false;
	}
    }
    
    final public Vertex GetPointAt(double parameter,boolean forward) {
	Vertex vertex;
	Vertex nextVertex;
	if (forward) {
	    vertex = firstVertex;
	    nextVertex = secondVertex;
	}
	else {
	    vertex =lastVertex;
	    nextVertex = beforeVertex;
	}
	Vertex tempVertex;
	double len = vertex.Dist2(nextVertex);
	while (nextVertex instanceof LineVertex && len < parameter) {
	    tempVertex = vertex;
	    vertex = nextVertex;
	    nextVertex = ((LineVertex)nextVertex).GetOppositeVertex(tempVertex);
	    len += vertex.Dist2(nextVertex);
	}
	return nextVertex;
    }
	    
	
    final public boolean Branch() {
	return ( (firstVertex instanceof EndVertex || lastVertex instanceof EndVertex) &&
		 !(firstVertex instanceof EndVertex && lastVertex instanceof EndVertex) );
    }

    final public boolean OneEdgeBranch() {
	return (secondVertex.equals(lastVertex) && Branch());
    }

    final public boolean Loop() {
	return firstVertex.equals(lastVertex);
    }

    final public Curve GetCurve() {
	Curve curve = new Curve(new Sample());
	curve.AddPoint(firstVertex);	    
	Vertex vertex = firstVertex;
	Vertex nextVertex = secondVertex;
	Vertex tempVertex;
	while (!(nextVertex.equals(lastVertex))) {
	    curve.AddPoint(nextVertex);	    
	    tempVertex = vertex;
	    vertex = nextVertex;
	    nextVertex = ((LineVertex)nextVertex).GetOppositeVertex(tempVertex);
	}
	curve.AddPoint(lastVertex);	    
	return curve;
    }

    final public String toString() {
	return firstVertex.getClass() + " to " + lastVertex.getClass() + ",    \tl = " + length;
    }
}

    static public boolean Hole(Path path1,Path path2) {
	return ( !path1.Loop() && !path1.Loop() && !path1.equals(path2) && 
		 (path1.firstVertex.equals(path2.firstVertex) && path1.lastVertex.equals(path2.lastVertex)) ||
		 (path1.firstVertex.equals(path2.lastVertex) && path1.lastVertex.equals(path2.firstVertex)) );
    }

    static public boolean Hole(Path path1,Path path2,Path path3) {
	return ( !path1.Loop() && !path2.Loop() && !path3.Loop() && 
		 !path1.equals(path2) && !path1.equals(path3) && !path2.equals(path3) && 
		 ((path1.firstVertex.equals(path2.firstVertex) && 
		   (path1.lastVertex.equals(path3.firstVertex) && path2.lastVertex.equals(path3.lastVertex) ||
		    path1.lastVertex.equals(path3.lastVertex) && path2.lastVertex.equals(path3.firstVertex))) ||
		  (path1.firstVertex.equals(path2.lastVertex) && 
		   (path1.lastVertex.equals(path3.firstVertex) && path2.firstVertex.equals(path3.lastVertex) ||
		    path1.lastVertex.equals(path3.lastVertex) && path2.firstVertex.equals(path3.firstVertex)))) );
    }


final class Paths extends Object {
    Vector paths;
    
    public Paths(PrincipalCurveClass principalCurve) {
	for(int i=0;i<principalCurve.GetSize();i++)
	    principalCurve.GetVertexAt(i).label = 0;
	paths = new Vector();
	Vertex vertex;
	for(int i=0;i<principalCurve.GetSize();i++) {
	    vertex = principalCurve.GetVertexAt(i);
	    if (!(vertex instanceof LineVertex)) {
		Vertex [] neighbors = vertex.GetNeighbors();
		for(int l=0;l<neighbors.length;l++)
		    if (neighbors[l].label == 0)
			AddPath(new Path(vertex,neighbors[l]));
	    }
	}
	for(int i=0;i<principalCurve.GetSize();i++) {
	    vertex = principalCurve.GetVertexAt(i);
	    if (vertex instanceof LineVertex && vertex.label == 0) { // complete circle
		AddPath(new Path(vertex));
	    }
	}
    }
    
    private Paths() {
	paths = new Vector();
    }
    
    final public void AddPath(Path path) {
	for(int i=0;i<GetSize();i++)
	    if (path.equals(GetPathAt(i)))
		return;
	paths.addElement(path);
    }
    
    final public void DeletePath(Path path) {
	paths.removeElement(path);
    }
    
    final public void DeletePathAt(int index) {
	paths.removeElementAt(index);
    }
    
    final public int FindPath(Path path) {
	return paths.indexOf(path);
    }

    final public int GetSize() {
        return paths.size();
    }

    final public Path GetPathAt(int i) {
        return (Path)paths.elementAt(i);
    }

    final public Paths GetIncidentPaths(Vertex vertex) {
	Paths incidentPaths = new Paths();
 	for(int i=0;i<GetSize();i++)
	    if (GetPathAt(i).firstVertex.equals(vertex) || GetPathAt(i).lastVertex.equals(vertex))
		incidentPaths.AddPath(GetPathAt(i));
	return incidentPaths;
    }

    final public String toString() {
	String s = new String();
 	for(int i=0;i<GetSize();i++) {
 	    s += (i+1) + ": " + GetPathAt(i) + "\n";
	}
	return s;
    }
}
}

