package PrincipalCurve.Vertex;

import java.awt.*;
import java.util.*;
import LinearAlgebra.*;
import PrincipalCurve.Vertex.Elements.*;

final public class StarOfManyVertex extends Vertex implements HasOneEdge,JoinVertex {
    MiddleVertexOfMany middleVertex;
    EndVertexOfTwo [] endVertexs;
    int degree;

    // for PrincipalCurveClass.JoinVertices()
    public StarOfManyVertex(Vertex vertex1,Vertex vertex2) {
	super(vertex1);
	degree = vertex1.GetDegree() + vertex2.GetDegree();
	endVertexs = new EndVertexOfTwo[degree];
	int d = 0;
	for(int i=0;i<vertex1.GetElementVertices().length;i++)
	    if (vertex1.GetElementVertices()[i] instanceof EndVertexOfTwo)
		endVertexs[d++] = (EndVertexOfTwo)vertex1.GetElementVertices()[i];
	for(int i=0;i<vertex2.GetElementVertices().length;i++)
	    if (vertex2.GetElementVertices()[i] instanceof EndVertexOfTwo)
		endVertexs[d++] = (EndVertexOfTwo)vertex2.GetElementVertices()[i];
	middleVertex = new MiddleVertexOfMany(endVertexs);
    }

    // for Degrade()
    public StarOfManyVertex(StarOfManyVertex starOfManyVertex,Vertex neighbor) {
	super(starOfManyVertex);
	degree = starOfManyVertex.GetDegree() - 1;
	endVertexs = new EndVertexOfTwo[degree];
	int d = 0,i;
	for(i=0;i<degree;i++,d++) {
	    if (i == d && neighbor.equals(starOfManyVertex.GetEdge(i).GetVertex2()))
		d++;
	    endVertexs[i] = starOfManyVertex.endVertexs[d];
	}
	if (d == i)
	    throw(new ArithmeticException("NOT NEIGHBOR!!" +
					  "\nStarOfManyVertex:\n" + starOfManyVertex + 
					  "\nneighbor:\n" + neighbor + "\n"));	    
	middleVertex = new MiddleVertexOfMany(endVertexs);
    }

    final public Vertex GetVertex() {
	return this;
    }

    final public EdgeDirection GetEdgeIndex(Vertex vertex) {
	return middleVertex.GetEdgeIndex(vertex);
    }
    
    final public void Maintain(Vertex neighbor) {
	for(int i=0;i<degree;i++) {
	    if (neighbor.equals(endVertexs[i].GetEdge1().GetVertex2())) {
		endVertexs[i] = MaintainEndVertex(neighbor);
		return;
	    }
	}
	throw(new ArithmeticException("NOT NEIGHBOR!!" +
				      "\nStarOfManyVertex:\n" + this + 
				      "\nneighbor:\n" + neighbor + "\n"));	    	
    }

    final public Vertex Degrade(Vertex neighbor) {
	if (degree > 3)
	    return new StarOfManyVertex(this,neighbor);
  	else
	    return new LineVertex(this,neighbor);
    }

    final public Edge GetEdge(int i) {
	return endVertexs[i].GetEdge1();
    }

    final public int GetEdgeIndex(int i) {
	return endVertexs[i].GetEdgeIndex1();
    }

    final public int GetDegree() { 
	return degree;
    }

    final public Edge [] GetEdges() { 
	Edge [] edges = new Edge[degree];
	for(int i=0;i<degree;i++)
	    edges[i] = GetEdge(i);
	return edges;
    }

    final public Vertex [] GetNeighbors() { 
	Vertex [] neighbors = new Vertex[degree];
	for(int i=0;i<degree;i++)
	    neighbors[i] = GetEdge(i).GetVertex2();
	return neighbors;
    }

    final public int [] GetEdgeIndexes() { 
	int [] edgeIndexes = new int[degree];
	for(int i=0;i<degree;i++)
	    edgeIndexes[i] = GetEdgeIndex(i);
	return edgeIndexes;
    }
    
    final public int [] GetNeighborIndexes() { 
	int [] neighborIndexes = new int[degree];
	for(int i=0;i<degree;i++)
	    neighborIndexes[i] = GetEdge(i).GetVertexIndex2();
	return neighborIndexes;
    }

    final public ElementVertex [] GetElementVertices() { 
	ElementVertex [] elementVertices = new ElementVertex[degree+1];
	elementVertices[0] = middleVertex;
	for(int i=0;i<degree;i++)
	    elementVertices[i+1] = endVertexs[i];
	return elementVertices;
    }

    final public ElementVertex GetMainElementVertex() { 
	return middleVertex;
    }

    final public double GetPenalty() {
	double penalty = 0.0;
	for(int i=0;i<degree;i++) {
	    try {
		penalty += ((EndVertexOfThree)endVertexs[i]).GetAnglePenalty();
	    }
	    catch (ClassCastException e) {
	    }
	    penalty += endVertexs[i].GetLengthPenalty();
	}
	return penalty;
    }

    final public NumeratorAndDenominator GetNumeratorAndDenominatorForPenalty() {
	NumeratorAndDenominator nd = new NumeratorAndDenominator(this);
	for(int i=0;i<degree;i++) {
	    try {
		nd.AddEqual(((EndVertexOfThree)endVertexs[i]).GetNumeratorAndDenominatorForAnglePenalty());
	    }
	    catch (ClassCastException e) {
	    }
	    nd.AddEqual(endVertexs[i].GetNumeratorAndDenominatorForLengthPenalty());
	}
	return nd;
    }

    final public Vertex Restructure() {
	return this;
    }

    final public String toString() {
	String s = "StarOfManyVertex(" + degree + "):\t";
	s += super.toString() + "\t";
	s += middleVertex.toString();
	for(int i=0;i<degree;i++)
	    s += endVertexs[i].toString();
	return s;
    }    
}
